from abc import ABCMeta, ABC

import psycopg2


class AbstractDataBase(ABC, metaclass=ABCMeta):

    def __query(self, callback, response_limit: int):
        """Выполнение запроса по получению"""

        def to_dict(item):
            if item:
                return dict(zip(columns, item))
            return {}

        with self.conn.cursor() as cursor:
            callback(cursor)
            columns = [col.name for col in cursor.description or []]
            if response_limit == 1:
                try:
                    response = to_dict(cursor.fetchone())
                except Exception as e:
                    return None
            elif response_limit <= 0:
                response = list(map(to_dict, cursor.fetchall()))
            else:
                response = list(map(to_dict, cursor.fetchmany(response_limit)))
            return response

    @staticmethod
    def error_to_sql(value):
        if value is None:
            return 'null'
        if not isinstance(value, str):
            return str(value)
        return f"'{value}'"

    @staticmethod
    def func_name_with_attrs_to_sql(func_name: str, *args):
        return f"{func_name}({','.join(map(DataBase.error_to_sql, args))})"

    @staticmethod
    def __create_query_execute(cursor, prefix: str, func_name: str, *args):
        proc_param = ",".join(str('%s') for _ in args)

        try:
            return cursor.execute(f"{prefix} {func_name}({proc_param})", args)
        except Exception as e:
            raise ValueError(
                f"""Ошибка с параметрами в {prefix} {DataBase.func_name_with_attrs_to_sql(func_name, *args)}
                \n {e}""")

    def function(self, *args, func_name: str, response_limit: int = -1, aggregate='*'):
        return self.__query(
            lambda cursor: self.__create_query_execute(cursor, f'SELECT {aggregate} FROM ', func_name, *args),
            response_limit)

    def procedure(self, *args, func_name: str):
        def call_procedure(cursor):
            self.__create_query_execute(cursor, 'call ', func_name, *args)
            self.conn.commit()

        return self.__query(call_procedure, response_limit=1)


class DataBase(AbstractDataBase):
    def __init__(self,
                 db_name: str,
                 user: str, password: str,
                 host: str = 'localhost', port: int = 5432):
        """
        Управление базой данных
        :param db_name: Название бд
        :param user: Имя пользователя
        :param password: Пароль пользователя
        :param host: Хост
        :param port: Порт
        """
        self.db_name = db_name
        self.user = user
        self.password = password
        self.host = host
        self.port = port

    def __enter__(self):
        assert all([self.db_name, self.user, self.password, self.host,
                    self.port]), 'Указаны не все данные для подключения к бд'
        try:
            self.conn = psycopg2.connect(dbname=self.db_name, user=self.user,
                                         password=self.password, host=self.host, port=self.port)
        except Exception as e:
            print("Ошибка при подключении к БД", '\n', e)
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.conn.close()



from datetime import datetime
from typing import List

from pydantic import BaseModel


class CustomField(BaseModel):
    """Базовый класс для кастомных типов данных"""

    @classmethod
    def get_validators(cls):
        yield cls.validate

    @classmethod
    def validate(cls, value, values, config, field):
        """Обязательная функция для валидации"""
        return value


class Const(CustomField):
    """Неизменяемое значение"""

    @classmethod
    def validate(cls, value, values, config, field):
        return field.default


class FormatDate(CustomField):
    """Поле для форматирования даты из строки"""
    _input_format: List[str]
    _output_format: str
    _ERROR = "Не правильный формат даты"

    @classmethod
    def validate(cls, value, values, config, field):
        print(cls._ERROR)
        if not value:
            return None
        for frmt in cls._input_format:
            try:
                dt = datetime.strptime(value, frmt)
            except Exception as e:
                pass
            else:
                try:
                    return dt.strftime(cls._output_format)
                except Exception:
                    raise ValueError(cls._ERROR)
        raise ValueError(cls._ERROR)



from django.db import connections

from controller.core.db import AbstractDataBase


class DjangoDataBase(AbstractDataBase):

    def __init__(self, db_name: str = 'default', *args):
        super(DjangoDataBase, self).__init__(*args)
        self.db_name = db_name

    def __enter__(self):
        self.conn = connections[self.db_name]
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        pass



from typing import Optional, Any

from controller.core.models import Model
from controller.django.fields import AuthUser


class DjangoModelMixin(Model):
    """Позволяет принимать в себя request, чтобы спарсить данные с запроса"""
    _request: Optional[Any] = None

    def __init__(self, **kwargs):
        def update_data(items):
            assert getattr(items, 'get')
            for key in items:
                if key.endswith('[]'):
                    data[key.replace('[]', '')] = items.getlist(key)
                else:
                    data[key] = items.get(key)

        data = {}
        request = kwargs.pop('request', None)
        if request:
            items = request.GET if request.method == 'GET' else request.data
            update_data(items)
        # В первую очередь идут параметры с запроса
        # Кваргсами есть возможность переопределить поля
        data.update(kwargs)
        # Невозможно переопределить AuthUser
        if request:
            for key in self.get_list_for_auth_user():
                data[key] = request.user.id if request.user.is_authenticated else None
        super(DjangoModelMixin, self).__init__(**data)

    def get_list_for_auth_user(self):
        # Примитивно, но работает
        return list(
            filter(lambda item: self.__annotations__[item] in [AuthUser, Optional[AuthUser]],
                   self.__annotations__)
        )



import re
from datetime import datetime
from typing import Any, Union

from pydantic import BaseModel, BaseConfig

from controller.core.db import DataBase
from controller.core.functions import count


def log(func):
    def wrapper(self, *args, **kwargs):
        self.send_log(func, args, kwargs)
        response = func(self, *args, **kwargs)
        return response

    return wrapper


class Model(BaseModel):
    _db: DataBase
    _func_name: str = None

    class Config(BaseConfig):
        use_enum_values = True
        validate_all = True
        validate_assignment = True
        error_msg_templates = {
            'type_error.none.not_allowed': 'Поле не может быть пустым',
            'value_error.missing': "Поле обязательное",
            'value_error.const': "Заданное значение не входит в список разрешенных для этого поля"
        }

    def __str__(self):
        return f"{self.__class__.__name__}({super(Model, self).__str__()})"

    @property
    def sql_func(self):
        return f"{self.func_name}({','.join(map(DataBase.error_to_sql, self.dict().values()))})"

    @staticmethod
    def camel_to_snake(full_name):
        def to_snake(name):
            return re.sub(r'(?<!^)(?=[A-Z])', '_', name)

        return '.'.join(map(to_snake, full_name.split('.'))).lower()

    @property
    def func_name(self):
        return self._func_name or self.camel_to_snake(self.__class__.__qualname__)

    @staticmethod
    def catch_response(data: Any):
        """Обработка ответа"""
        return data

    def send_log(self, func, args, kwargs):
        now = datetime.now().strftime('%d.%m.%Y %H:%M:%S')
        print(f'{now}: Вызов {self.func_name} {func.__name__}({args};{kwargs}) с параметрами: {self.dict()}')


class FunctionModel(Model):

    def __len__(self):
        return self.length()

    @log
    def get(self, limit: int = 1) -> Union[dict, list]:
        """Получение n элементов"""
        with self._db as db:
            return self.catch_response(
                db.function(*self.dict().values(), func_name=self.func_name, response_limit=limit))

    @log
    def all(self):
        """Получение всех элементов"""
        with self._db as db:
            return self.catch_response(db.function(*self.dict().values(), func_name=self.func_name, response_limit=-1))

    @log
    def length(self) -> int:
        """Получение количество строк"""
        with self._db as db:
            return db.function(*self.dict().values(),
                               func_name=self.func_name, response_limit=1,
                               aggregate=count()).get('count', 0)


class ProcedureModel(Model):
    @log
    def execute(self):
        """Выполнить"""
        with self._db as db:
            return self.catch_response(db.procedure(*self.dict().values(), func_name=self.func_name))



from django.http import JsonResponse
from django.utils.deprecation import MiddlewareMixin
from pydantic import ValidationError


class ErrorMiddleware(MiddlewareMixin):

    def process_exception(self, request, exception):
        if isinstance(exception, ValidationError):
            return JsonResponse({'errors': exception.errors()}, status=500)
        else:
            raise exception
